#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

################################################
#############   MACHINE LEARNING   #############
##############     -  2016 -    ################
################################################

"""
Exercise 5: CNN face segmentation on
            Caltech faces images.

Call without arguments: 
 > python faceseg.py

Contact: Adrian Ulges (adrian.ulges@hs-rm.de)
"""
import os
import tensorflow as tf
import random
import cv2
from numpy import genfromtxt
import numpy as np
import time

# this code does not the (Caltech) input images
# at full resolution but scales to WIDTH x HEIGHT
# pixels *before* processing them.
HEIGHT = 148
WIDTH = 224

# the CNN's learning rate
LRATE = 2E-3

# the number of images per minibatch
BATCHSIZE_TRAIN = 15
BATCHSIZE_TEST  = 30


def read_data_from_list(filename, gt):
    """
    reads a list of images from a textfile (e.g., 'train.txt'), 
    rescales the input images, and produces binary  ground truth
    segmentation masks with rectangular face regions marked white.
    """
    imgs = []
    masks = []
    masks_mini = []

    with open(filename) as stream:

        lines = stream.readlines()

        for l in lines:
            
            # read input image and ground truth
            f = l.rstrip("\n")
            imgno = int(l[-9:-5])
            img = cv2.imread(f)
            x = gt[:,imgno-1]
            print f

            # estimate face region
            x0 = int(0.5*(x[0]+x[2]))
            x1 = int(0.5*(x[4]+x[6]))
            y0 = int(0.5*(x[3]+x[5]))
            y1 = int(0.5*(x[1]+x[7]))

            # reduce the face region to a more
            # reasonable size
            dx = x1-x0
            dy = y1-y0
            x0 = x0 + dx / 10  # remove 10% left 
            x1 = x1 - dx / 10  # remove 10% right
            y0 = y0 + dy / 4   # remove 25% top
            y1 = y1 - dy / 10  # remove 10% bottom
            mask = np.zeros((img.shape[0],img.shape[1]), dtype='uint8')
            mask[y0:y1,x0:x1] = 255

            # scale down image+mask to WIDTH x HEIGHT pixels
            img = cv2.resize(img, (WIDTH,HEIGHT))
            mask = cv2.resize(mask, (WIDTH,HEIGHT))
            mask_mini = cv2.resize(mask, (WIDTH/4,HEIGHT/4))
            mask_mini.shape = (HEIGHT/4,WIDTH/4,1)

            # scale pixel range from [0,255] to [0,1]
            mask_mini = mask_mini.astype('float') / 255.

            imgs.append(img)
            masks.append(mask)
            masks_mini.append(mask_mini)

    return imgs, masks, masks_mini

def read_ground_truth():
    """ reads ground truth face coordinates from csv file """
    data = genfromtxt('caltech/gt.csv', delimiter=',')
    return data


"""
    CNN network structure
"""

def weight_variable(shape):
  initial = tf.truncated_normal(shape, stddev=0.1)
  return tf.Variable(initial)

def bias_variable(shape):
  initial = tf.constant(0.1, shape=shape)
  return tf.Variable(initial)

def conv2d(x, W):
  return tf.nn.conv2d(x, W, strides=[1, 1, 1, 1], padding='SAME')

def max_pool_2x2(x):
  return tf.nn.max_pool(x, ksize=[1, 2, 2, 1],
                        strides=[1, 2, 2, 1], padding='SAME')


def CNN(session):

    x_ = tf.placeholder(tf.float32, shape=[None, HEIGHT, WIDTH, 3])
    y_ = tf.placeholder(tf.float32, shape=[None, HEIGHT/4, WIDTH/4, 1])
    
    W_conv1 = weight_variable([5, 5, 3, 16])
    b_conv1 = bias_variable([16])

    h_conv1 = tf.nn.relu(conv2d(x_, W_conv1) + b_conv1)
    h_pool1 = max_pool_2x2(h_conv1)

    W_conv2 = weight_variable([5, 5, 16, 32])
    b_conv2 = bias_variable([32])
    h_conv2 = tf.nn.relu(conv2d(h_pool1, W_conv2) + b_conv2)
    # This layer is 1/16 of the original input image: 37x56x32
    h_pool2 = max_pool_2x2(h_conv2)

    # 1x1 convolution -> creates one image out of 32 feature maps
    W_1x1 = weight_variable([1, 1, 32, 1])
    b_1x1 = bias_variable([1])
    logits = conv2d(h_pool2, W_1x1) + b_1x1
    y = tf.nn.sigmoid(logits)

    # loss function for training (defined on the *logits*, *not* on y!)
    cross_entropy = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(logits, y_))
    train_step = tf.train.AdamOptimizer(LRATE).minimize(cross_entropy)

    return (logits,y,train_step,x_,y_)


"""
   data batching
"""

def next_batch(data, size):
    (imgs, masks, masks_mini) = data
    n = len(imgs)
    idx = random.sample(range(n), size)
    x = np.array(imgs)[idx]
    y = np.array(masks_mini)[idx]
    return x,y

"""
   main routine: trains and tests the CNN
"""

def train_and_eval(data_train, data_test, model, session):

  (logits,y,train_step,x_,y_) = model

  tf.initialize_all_variables().run()

  before = time.time()

  # keep test batch fixed to track
  # progress over learning iterations
  xtest,ytest = next_batch(data_test, BATCHSIZE_TEST)

  for _ in range(1000):

      # run a training iteration
      print _
      xtrain,ytrain = next_batch(data_train, BATCHSIZE_TRAIN)
      session.run(train_step, feed_dict={x_:xtrain, y_:ytrain})

      # check: pixel-wise difference beteen CNN output
      #        and ground truth (Jaccard not implemented here)
      # ... once for training data
      yres = session.run(y, feed_dict={x_:xtrain})
      error_train = np.mean(np.abs(yres-ytrain))
      # ... once fot testing data
      yres = session.run(y, feed_dict={x_:xtest})
      error_test = np.mean(np.abs(yres-ytest))
      print "Error: %.2f (train)  %.2f (test)" %(error_train, error_test)

      # track a sample mask over the training iterations
      mask = ((yres[0][:,:,0])*255).astype('uint8')
      cv2.imwrite('mask-%.4d.png' %(_), mask)

  after = time.time()
  print 'Training took %s seconds' %(after-before)

if __name__ == "__main__":

    # reading input data
    gt = read_ground_truth()
    data_train = read_data_from_list('train.txt', gt)
    data_test  = read_data_from_list('test.txt', gt)

    session = tf.InteractiveSession()
    model = CNN(session)

    train_and_eval(data_train, data_test, model, session)
